# (c) 2020 Cluso99 (Ray Rodrick)
# License: MIT
#
# RR20200122 v001  commence
# RR20200123 v009  basics wkg - single block
# RR20200124 v010  use start/end comments to define a block (no longer each line)
#            v011  process each block
# RR20200125 v018  basics wkg

# scan thru' spin2 file and calculate skip bitmap for execf bytecodes

from time import time
from datetime import datetime, date, timedelta
import os

posn = 0

def main_process():
    global posn

    path_name       = 'C:/P2/Spin2/'
    fname_in        = 'source.spin2'
    fname_out       = '_execf'

    #==============================================================================#
    # Read source, extract / process, write out all relevant code blocks           #
    #==============================================================================#
    fnamei    = path_name + fname_in
    lst       = fnamei.rsplit('.',1)
    fnameo    = lst[0] + fname_out + '.' + lst[1]

    posn = 0
    lines_out = []
    lines_out.append("'' " + fnamei + " --> " + fnameo)                             # add filenames to output file

    lines_in  = read_file(fnamei)                                                   # read source


    while posn < len(lines_in):
        lines = get_block(lines_in)                                                 # return next code block
        for line in lines:                                                          # add source block as comment to output file
            lines_out.append("''" + line)
        lines = (process_block(lines))                                              # return processed block of code
        lines_out.extend(lines)                                                     # add processed block to output file
        lines_out.append("''")                                                      # add blank comment to output file

    write_file(fnameo, lines_out)                                                   # write out all processed blocks


#==============================================================================#
# Get next block of code                                                       #
#==============================================================================#
def get_block(lines_in):
    global posn
    lines = []
    found = False
    for i in range(posn,len(lines_in)):
        posn = i+1                                                                  # save next position
        line = lines_in[i]
        if not found:
            if line[0:5] == "'~---":                                                # start of block?
                found = True                                                        #   y: set flag 
                lines.append(line)                                                  #   y: and copy
        else:
            lines.append(line)                                                      # copy
            if line[0:5] == "'~~--":                                                # end of block?
                break                                                               #   y: exit 'for loop
    return(lines)


#==============================================================================#
# Process a code block                                                         #
#==============================================================================#
def process_block(lines_in):
    max_tables = 52                                                                 # max no of skip tables
    lines     = []                                                                  # list of processed lines
    labels    = []                                                                  # list of instruction lables
    instrs    = []                                                                  # list of instructions (placeholder)
    tables    = []                                                                  # list of skip tables
    tblhdr    = []                                                                  # list of skip tables header character a-z,A-Z
    for t in range(0,max_tables):                                                   # max table entries in a table
        tables.append('')                                                           # clear out table list
        tblhdr.append('')                                                           # clear out table header character

    #==============================================================================#
    # Process each code line                                                       #
    #==============================================================================#
    for line in lines_in:
        line = line.replace('\t','    ')                                            # replace <tab> with 4 spaces
        if line.startswith("'~"):
            pass                                                                    # ignore start/end block comments
        elif (len(line.strip()) == 0):
            pass                                                                    # ignore blank lines
        elif (line.strip().find("'") < 1):
            pass                                                                    # ignore comment only lines
        elif (line.find("'") > -1):                                                 # comment?
            char0 = line[0]
            if char0.isalpha():                                                     # label? (first char alpha)
                label = line[:line.find(" ")]                                       # get the label
                instr = line[len(label):line.find("'")].strip()                     # get the instruction
            elif char0 == "'":                                                      # comment line?
                label = ''                                                          #   no label
                instr = ''                                                          #   no instruction
            else:                                                                   # no label
                label = ''                                                          #   no label
                instr = line[:line.find("'")].strip()                               #   get the instruction
            comment =  line[line.find("'")+1:]                                      # get the comment

#            print(label +' : ' + instr +' : ' + comment)

            #==============================================================================#
            # now create the table from the comment info...                                #
            #==============================================================================#
            if len(instr) == 0:
                lines.append('** missing instruction: ' + line)

            labels.append(label)                                                    # add the label to the list
            instrs.append(instr)                                                    # add the instruction to the list
            comment = comment.ljust(max_tables)                                     # make length>=52 for a-z,A-Z columns

            for t in range(0,max_tables):                                           # create table entries for this line
                ch = comment[t]                                                     # get a table entry char
                if ch.isalpha():                                                    # a-z,A-Z ?
                    if tblhdr[t] == '':                                             #   first time?
                        tblhdr[t] = ch                                              #      y: so copy
                        tables[t] = '0' + tables[t]                                 #      y: add '0' to respective table array
                    else:                                                           #
                        if tblhdr[t] == ch:                                         #      if no, validate
                            tables[t] = '0' + tables[t]                             #         y: add '0' to respective table array
                        else:                                                       #
                            tblhdr[t] = '?'                                         #         n: mark as invalid table
                            tables[t] = '?' + tables[t]                             #         n: add '?' to respective table array
                elif ch == '|':
                    tables[t] = '1' + tables[t]                                     # add '1' to respective table array
                elif ch == ' ':
                    tables[t] = ' ' + tables[t]                                     # add ' ' to respective table array
                else:
                    tblhdr[t] = '?'                                                 # invalid: mark as invalid table
                    tables[t] = '?' + tables[t]                                     # invalid: add '?' to respective table array
#            print(tblhdr[t] +': ' + tables[t])
        else:
            pass                                                                    # ??? how to validate if no comment???

#    print('-----')
#    print(len(tables))

    #==============================================================================#
    # now process valid entries...                                                 #
    #==============================================================================#
    for t in range(0,len(tables)):                                                  # copy valid entries... 
        if tblhdr[t].isalpha():
#            print(str(t) + ': '+ tblhdr[t] + ': ' + tables[t] +'<')
            line = process_entry(labels, tblhdr[t], tables[t])
            lines.append(line)
        elif tblhdr[t] == '?':                                                      # terminate once invalid entry found ie end of list
            break

    return(lines)


#==============================================================================#
# process a valid entry...                                                     #
#==============================================================================#
def process_entry(labels, ch, mask):                                                # process valid entries
    n     = len(mask) - len(mask.rstrip())                                          # calculate index to correct label
    label = labels[n]
    txt   = label + '_skip_' + ch 
    line  = txt.ljust(24) + '= ' + label.ljust(16) + '| '

    msk = mask.strip()
    if len(msk) > 22:                                                               # > 22 bits? OK if extras are all 0's             
        tmp = msk[:-22]
        tmp = tmp.replace('0','')
        if len(tmp) > 0:        
            line = '** mask > 22: ' + line

    for i in range(0,len(msk)):                                                     # check all valid 0/1 ?
        if (msk[i] == '0') or (msk[i] == '1') :
            pass
        else:
            line = '** mask invalid: ' + line
            break   
    
    if len(msk) > 22:
        txt   = "%" + msk[:-22] + '_' + msk[-22:] + " << 10  '"
    else:
        txt   = "%" + msk
        txt   = txt.rjust(23)  + " << 10  '"

    line  = line + txt

    return(line)

#------------------------------------------------------------------------------#
def write_file(fname,lines):                    # write file from list
    print('i) Writing:    ' + fname)
    with open(fname, 'w', encoding='utf-8') as f:
        for line in lines:
            f.write(line + '\n')
        f.close()
#------------------------------------------------------------------------------#
def read_file(fname):                           # read file into list
    print('i) Reading:    ' + fname)
    with open(fname, 'r', encoding='utf-8') as f:
        line = f.read()
        lines = line.split('\n')                # split into lines
        f.close()
        return(lines)
#------------------------------------------------------------------------------#
def startprog(txt):                             # start program message
    print('i) Program <' + txt + '> begin... ' + datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
    return(time())                              # start time
#------------------------------------------------------------------------------#
def endprog(starttime):                         # end program message
    print('i) Program end. (' + str(round((time() - starttime)/60,1)) +'m)  ' + datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
#------------------------------------------------------------------------------#

#------------------------------------------------------------------------------#
starttime = startprog('P2-skip-tables')
main_process()                              # go do it
endprog(starttime)
#------------------------------------------------------------------------------#
